#!/usr/bin/python3

import rospy
import numpy as np
import sys
from dji_sdk.srv import SDKControlAuthority
from dji_sdk.srv import MissionWpUpload
from dji_sdk.srv import MissionWpAction
from dji_sdk.srv import MissionWpActionRequest
from dji_sdk.srv import DroneTaskControl
from dji_sdk.srv import DroneTaskControlRequest
from dji_sdk.srv import MissionWpUploadRequest

from dji_sdk.msg import MissionWaypoint
from dji_sdk.msg import MissionWaypointTask
from dji_sdk.msg import MissionWaypoint
from dji_sdk.msg import MissionWaypointAction

from dji_sdk.srv import WaypointControl
from dji_sdk.srv import WaypointControlResponse

class WaypointDriver():

    def setupWaypoints(self):

        # wait for parameters to show up
        while not rospy.has_param("/robot_list/robot_"+str(self.droneId)+"/numWaypoints"):
            rospy.loginfo("waiting for waypoint parameters...")

        missionTask = MissionWaypointTask()
        missionTask.velocity_range     = np.float32(2.0)
        missionTask.idle_velocity      = np.float32(self.baseSpeed)
        missionTask.action_on_finish   = np.uint8(MissionWaypointTask.FINISH_NO_ACTION)
        missionTask.mission_exec_times = np.uint8(1)
        missionTask.yaw_mode		    = np.uint8(MissionWaypointTask.YAW_MODE_RC)
        missionTask.trace_mode	        = np.uint8(MissionWaypointTask.TRACE_COORDINATED)
        missionTask.action_on_rc_lost  = np.uint8(MissionWaypointTask.ACTION_AUTO)
        missionTask.gimbal_pitch_mode  = np.uint8(MissionWaypointTask.GIMBAL_PITCH_FREE)

        numWaypoints = rospy.get_param("/robot_list/robot_"+str(self.droneId)+"/numWaypoints")
        waypointList = []

        for i in range(numWaypoints):
            waypt = rospy.get_param("/robot_list/robot_"+str(self.droneId)+"/waypoints/waypoint_" + str(i))
            rospy.loginfo("Waypoint read as (Ind, Lat, Lon, Alt, Yaw): {:.0f} {:.3f} {:.3f} {:.3f} {:.3f}".format(waypt[0], waypt[1], waypt[2], waypt[3], waypt[4]))
            missionWaypoint = MissionWaypoint()

            # specific settings
            missionWaypoint.latitude = np.float64(waypt[1])
            missionWaypoint.longitude = np.float64(waypt[2])
            missionWaypoint.altitude = np.float32(waypt[3])
            missionWaypoint.target_yaw = np.int16(0)
            # missionWaypoint.target_yaw = np.int16(waypt[4])

            # default settings
            missionWaypoint.damping_distance    = np.float32(1.0)
            missionWaypoint.target_gimbal_pitch = np.int16(0)
            missionWaypoint.turn_mode           = np.uint8(1)
            missionWaypoint.has_action          = np.uint8(0)
            missionWaypoint.action_time_limit   = np.uint16(0)

            missionWaypoint.waypoint_action = MissionWaypointAction()
            missionWaypoint.waypoint_action.action_repeat = np.uint8(0)
            missionWaypoint.waypoint_action.command_list = [np.uint8(0)]*16
            missionWaypoint.waypoint_action.command_parameter = [np.uint16(0)]*16

            waypointList.append(missionWaypoint)

        missionTask.mission_waypoint = waypointList
        rospy.loginfo("sending waypoint upload command")
        rospy.loginfo(sys.getsizeof(missionTask))
        # waypointUpload = MissionWpUpload()
        waypointUpload = MissionWpUploadRequest()
        waypointUpload.waypoint_task = missionTask
        response = self.missionWaypointUploadProxy(waypointUpload)
        rospy.loginfo(response)

    def startWaypoints(self):
        waypointAction = MissionWpActionRequest()
        waypointAction.action = MissionWpActionRequest.ACTION_START
        response = self.missionWaypointActionProxy(waypointAction)
        return response

    def stopWaypoints(self):
        waypointAction = MissionWpActionRequest()
        waypointAction.action = MissionWpActionRequest.ACTION_STOP
        response = self.missionWaypointActionProxy(waypointAction)
        return response

    def land(self):
        flightTaskControl = DroneTaskControlRequest()
        flightTaskControl.task = DroneTaskControlRequest.TASK_LAND
        response = self.droneTaskControlProxy(flightTaskControl)
        return response

    def takeoff(self):
        flightTaskControl = DroneTaskControlRequest()
        flightTaskControl.task = DroneTaskControlRequest.TASK_TAKEOFF
        response = self.droneTaskControlProxy(flightTaskControl)
        return response

    def goHome(self):
        flightTaskControl = DroneTaskControlRequest()
        flightTaskControl.task = DroneTaskControlRequest.TASK_GOHOME
        response = self.droneTaskControlProxy(flightTaskControl)
        return response

    def controlService(self, req):
        if (req.data == 0):
            # takeoff
            rospy.loginfo("recieved takeoff command")
            response = self.takeoff()
            rospy.loginfo(response)
        elif (req.data == 1):
            # land
            rospy.loginfo("recieved land command")
            response = self.goHome()
            rospy.loginfo(response)
            response = self.land()
            rospy.loginfo(response)
        elif (req.data == 2):
            # land
            rospy.loginfo("recieved start waypoints command")
            response = self.setupWaypoints()
            rospy.loginfo(response)
            # response = self.takeoff()
            # rospy.loginfo(response)
            response = self.startWaypoints()
            rospy.loginfo(response)
        elif (req.data == 3):
            # land
            rospy.loginfo("recieved stop waypoints command")
            response = self.stopWaypoints()
            rospy.loginfo(response)
            response = self.goHome()
            rospy.loginfo(response)
            response = self.land()
            rospy.loginfo(response)
        else:
            rospy.loginfo("unknown control request")

        return WaypointControlResponse(result = True)

    def __init__(self):
        rospy.init_node('waypointDriver')

        # make service proxies for all services we need
        rospy.loginfo("waiting for services...")
        rospy.wait_for_service('/dji_sdk/sdk_control_authority')
        rospy.wait_for_service('/dji_sdk/mission_waypoint_upload')
        rospy.wait_for_service('/dji_sdk/mission_waypoint_action')
        rospy.wait_for_service('/dji_sdk/drone_task_control')
        rospy.loginfo("all services ready!")

        rospy.loginfo("waiting for parameters...")

        while not rospy.has_param("~drone_id"):
            rospy.loginfo("waiting for drone_id param...")
        self.droneId = rospy.get_param("~drone_id")

        while not rospy.has_param("~base_speed"):
            rospy.loginfo("waiting for base_speed param...")
        self.baseSpeed = rospy.get_param("~base_speed")

        rospy.loginfo("all parameters recieved!")

        self.sdkControlProxy = rospy.ServiceProxy('/dji_sdk/sdk_control_authority', SDKControlAuthority)
        self.missionWaypointUploadProxy = rospy.ServiceProxy('/dji_sdk/mission_waypoint_upload', MissionWpUpload)
        self.missionWaypointActionProxy = rospy.ServiceProxy('/dji_sdk/mission_waypoint_action', MissionWpAction)
        self.droneTaskControlProxy = rospy.ServiceProxy('/dji_sdk/drone_task_control', DroneTaskControl)
        rospy.Service('waypoint_control', WaypointControl, self.controlService)

        rospy.loginfo("requesting sdk control authority...")
        response = self.sdkControlProxy(np.uint8(1))
        rospy.loginfo(response)
        rospy.loginfo("obtained control authority.")

if __name__ == '__main__':
    wayptDriver = WaypointDriver()
    rospy.spin()
