#!/usr/bin/python3
from pyproj import Transformer
import numpy as np
import rospy
import math

def meters2lat_lon(meters_point):
    # returns point converted to lat lon from meters coordinate system
    transformer = Transformer.from_crs(3857, 4326)
    latlon_point = transformer.transform(meters_point[0], meters_point[1])
    return latlon_point

def lat_lon2meters(latlon_point):
    # returns point converted to meters from lat lon coordinate system
    transformer = Transformer.from_crs(4326, 3857)
    meters_point = transformer.transform(latlon_point[0], latlon_point[1])
    return meters_point

def normalizeAngle(angle):
    newAngle = angle%360
    newAngle = (newAngle + 360)%360
    if(newAngle >= 180):
        newAngle -= 360
    return newAngle

def point_rotation(origin, pt, ang):
    # returns the pt rotated about the origin by ang (ang in degrees)
    c = math.cos(math.radians(ang))
    s = math.sin(math.radians(ang))
    # translate to origin
    pt_temp = [pt[0] - origin[0], pt[1] - origin[1]]
    pt_spun = [ pt_temp[0]*c - pt_temp[1]*s, pt_temp[0]*s + pt_temp[1]*c ]
    # translate back to frame
    pt_spun = [pt_spun[0] + origin[0], pt_spun[1] + origin[1]]
    return pt_spun

def generateWaypoints():
    rospy.init_node("waypointGenerator")
    anchorPoint = rospy.get_param("~anchorPoint")
    pathPoints = rospy.get_param("~pathPoints")
    pathType = rospy.get_param("~pathType")
    baseAltitude = rospy.get_param("~altitude")
    droneNum = int(rospy.get_param("~droneNum"))
    spiralRadius = rospy.get_param("~spiralRadius")
    spiralHeight = rospy.get_param("~spiralHeight")
    spiralLoops = int(rospy.get_param("~spiralLoops"))
    altitudeIncrement = rospy.get_param("~altitudeIncrement") # how much to separate drones on z axis

    apMeters = lat_lon2meters(anchorPoint)
    tempDict = {"waypoints" : {}, "positions" : {}, "numWaypoints" : 0, "numPositions" : 0}
    waypointDict = {}

    # path type if
    if (pathType == "points" and len(pathPoints) != 0):

        for d in range(droneNum):
            waypointDict["robot_{}".format(d)] = {}
            waypointDict["robot_{}".format(d)] = {}
            waypointDict["robot_{}".format(d)]["waypoints"] = {}
            waypointDict["robot_{}".format(d)]["positions"] = {}

            # quick check if path points has levels
            try:
                for i, pt in enumerate(pathPoints[d]):
                # for i, offset in enumerate(waypointOffsets):
                    waypointLatLon = pt
                    heading = 0.0
                    latLonPt = lat_lon2meters(pt)
                    positionXY = [latLonPt[0] - apMeters[0], latLonPt[1] - apMeters[1]]

                    # waypoints
                    waypointDict["robot_{}".format(d)]["waypoints"].update({"waypoint_{}".format(i) : [i, waypointLatLon[0], waypointLatLon[1], baseAltitude + d*altitudeIncrement, heading]})
                    # positions
                    waypointDict["robot_{}".format(d)]["positions"].update({"position_{}".format(i) : [i, positionXY[0], positionXY[1], baseAltitude + d*altitudeIncrement, heading]})

                # do this betta
                waypointDict["robot_{}".format(d)]["numWaypoints"] = len(pathPoints[d])
                waypointDict["robot_{}".format(d)]["numPositions"] = len(pathPoints[d])
            except TypeError as e:
                print("path points are not ordered correctly")
                rospy.loginfo("path points are not ordered correctly")

    elif (pathType == "squares"):
        # musical drones trajectory is convient test case, but general path is possible
        # ring sizes
        big = 10.0
        med = 5.0
        sml = 2.0

        waypointOffsets = [
        [-big, -big],
        [big, -big],
        [big, big],
        [med, med],
        [-med, med],
        [-med, -med],
        [-sml, -sml],
        [sml, -sml],
        [sml, sml],
        [big, -big],
        ]

        for d in range(droneNum):
            # rospy.loginfo("blah!")
            # print(d)
            waypointDict["robot_{}".format(d)] = {}
            waypointDict["robot_{}".format(d)] = {}
            waypointDict["robot_{}".format(d)]["waypoints"] = {}
            waypointDict["robot_{}".format(d)]["positions"] = {}
            for i, offset in enumerate(waypointOffsets):
                waypointLatLon = meters2lat_lon([apMeters[0] + offset[0], apMeters[1] + offset[1]])
                waypointLatLonRev = meters2lat_lon([apMeters[0] - offset[0], apMeters[1] - offset[1]])
                heading = normalizeAngle(45*i)

                # waypoints
                waypointDict["robot_{}".format(d)]["waypoints"].update({"waypoint_{}".format(i) : [i, waypointLatLon[0], waypointLatLon[1], baseAltitude + d*waypointAltitudeIncrement, heading]})
                # positions
                waypointDict["robot_{}".format(d)]["positions"].update({"position_{}".format(i) : [i, offset[0], offset[1], baseAltitude + d*waypointAltitudeIncrement, heading]})

            # do this betta
            waypointDict["robot_{}".format(d)]["numWaypoints"] = len(waypointOffsets)
            waypointDict["robot_{}".format(d)]["numPositions"] = len(waypointOffsets)
    elif (pathType == "spiral"):
        # sprial path with heading pointing towards center of spiral
        theta = 0.0
        radians = np.pi*2.0*spiralLoops # total radians in spiral
        loopPoints = 20 # 20 steps per single loop
        totalPoints = loopPoints*spiralLoops
        thetaStep = radians/totalPoints
        heightStep = spiralHeight/totalPoints
        height = 0.0

        for d in range(droneNum):
            waypointDict["robot_{}".format(d)] = {}
            waypointDict["robot_{}".format(d)] = {}
            waypointDict["robot_{}".format(d)]["waypoints"] = {}
            waypointDict["robot_{}".format(d)]["positions"] = {}
            # spiral loop
            for i in range(totalPoints):
                offset = [spiralRadius*math.cos(theta), spiralRadius*math.sin(theta)]
                waypointLatLon = meters2lat_lon([apMeters[0] + offset[0], apMeters[1] + offset[1]])
                heading = normalizeAngle(math.degrees(-theta - (np.pi/2.0)))

                # waypoints
                waypointDict["robot_{}".format(d)]["waypoints"].update({"waypoint_{}".format(i) : [i, waypointLatLon[0], waypointLatLon[1], baseAltitude + height + d*altitudeIncrement, heading]})
                # positions
                waypointDict["robot_{}".format(d)]["positions"].update({"position_{}".format(i) : [i, offset[0], offset[1], baseAltitude + height + d*altitudeIncrement, heading]})
                theta = theta + thetaStep
                height = height + heightStep

            # do this betta
            waypointDict["robot_{}".format(d)]["numWaypoints"] = totalPoints
            waypointDict["robot_{}".format(d)]["numPositions"] = totalPoints
    rospy.set_param("/robot_list", waypointDict)
    return
    # print(waypointDict)

if __name__ == "__main__":
    generateWaypoints()
